#pragma once
#include <cstdint>
#include <algorithm>

#if __BYTE_ORDER != __LITTLE_ENDIAN
#error "TODO: support big endian"
#endif

typedef unsigned int uint;
typedef uint8_t byte;

namespace ADAPCM {

typedef uint16_t sample_t;
typedef int16_t coeff_t;
typedef int8_t delta8_t;

constexpr uint PAGE_SIZE = 512;
constexpr uint PAGE_HEADER_STEREO = 2*2*2;
constexpr uint PAGE_HEADER_MONO = 2*2*1;

inline uint bytes_per_block(bool stereo,uint samples_per_block) {
    if (stereo) return 1+samples_per_block;
    else        return 1+((samples_per_block+1)>>1);
}

inline uint blocks_per_page(bool stereo,uint samples_per_block) {
    uint header = stereo ? PAGE_HEADER_STEREO : PAGE_HEADER_MONO;
    return ((PAGE_SIZE-header) / bytes_per_block(stereo,samples_per_block));
}

inline uint samples_per_page(bool stereo,uint samples_per_block) {
    return 2 + blocks_per_page(stereo,samples_per_block) * samples_per_block;
}

inline byte getnib1(byte b) {
    return b >> 4;
}
inline byte getnib0(byte b) {
    return b & 0xf;
}
inline void setnib1(byte& b,uint nib) {
    b = (b&0x0f)+((nib&0x0f)<<4);
}
inline void setnib0(byte& b,uint nib) {
    b = (b&0xf0)+((nib&0x0f));
}


inline int32_t coeffmul(sample_t s,coeff_t c) {
    return (int32_t(s)-0x8000)*c;
}
constexpr coeff_t coefficents[16][2] = {
    {0,0},
    {2048,0},
    {2048,-2048},
    {2000,-1000},

    {256,0},
    {3420,-800},
    {1993,-1024},
    {3908,-4012},

    {2800,-2800},
    {1024,-123},
    {1600,-5},
    {4096,-1982},

    {2048,-500},
    {3438,-1637},
    {5000,-5000},
    {15000,-20000},
};

inline int32_t predict_sample(byte coeff_set, sample_t history1, sample_t history2) {
    return (coeffmul(history1,coefficents[coeff_set][0]) + coeffmul(history2,coefficents[coeff_set][1]) ) >> 11;
}

inline sample_t int32_to_sample_t(int32_t i) {
    return sample_t(0x8000+std::clamp(i,-0x8000,0x7FFF));
}

constexpr byte joint_upgrade_tbl[16] = {
    13,13,13,13,13,
    14,14,14,14,
    15,15,15,15,
    13,14,15
};

}