#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "adapcm.hpp"
#include "decode.hpp"
#include "encode.hpp"


using namespace ADAPCM;

int main(int argc,char **argv) {
    bool stereo;
    uint samples_per_block;
    FILE *infile,*outfile;

    if (argc!=5) goto print_usage;

         if (!strcmp(argv[2],"1")) stereo = false;
    else if (!strcmp(argv[2],"2")) stereo = true;
    else goto print_usage;

    samples_per_block = stereo ? 7 : 14;

    infile = fopen(argv[3],"rb");
    if (!infile) goto print_infile_error;
    outfile = fopen(argv[4],"wb");
    if (!outfile) goto print_outfile_error;

    if (!strcmp(argv[1],"d")) { // Decode
        byte inbuffer[PAGE_SIZE];
        uint pagesamples = samples_per_page(stereo,samples_per_block)*(stereo?2:1);
        sample_t outbuffer[pagesamples];
        scale_used = coeff_used = 0;
        memset(coeff_histogram,0,sizeof(coeff_histogram));
        while(fread(inbuffer,PAGE_SIZE,1,infile)) {
            if (!stereo) goto print_nyi;
            decode_page_stereo(inbuffer,outbuffer,samples_per_block);
            fwrite(outbuffer,2,pagesamples,outfile);
        }
        fclose(infile);
        fclose(outfile);
        printf("Scales used: %04X Coefficents used: %04X\n",scale_used,coeff_used);
        for (int i=0;i<16;i++) {
            printf("Coeff set %d used %llu times\n",i,coeff_histogram[i]);
        }
    } else if (!strcmp(argv[1],"e")) { // Encode
        uint pagesamples = samples_per_page(stereo,samples_per_block)*(stereo?2:1);
        sample_t inbuffer[pagesamples];
        byte outbuffer[PAGE_SIZE];
        int i = 0;
        while(fread(inbuffer,2,pagesamples,infile)) {
            if (!stereo) goto print_nyi;
            //printf("Processing page %d\n",i++);
            encode_page_stereo(inbuffer,outbuffer,samples_per_block);
            fwrite(outbuffer,PAGE_SIZE,1,outfile);
        }
        fclose(infile);
        fclose(outfile);
    } else goto print_usage;
    exit(0);
print_usage:
    printf("Usage: adapcm (e|d) (1|2) infile outfile\n");
    exit(-2);
print_nyi:
    printf("Not yet implemented...\n");
    exit(-3);
print_infile_error:
    printf("Error opening input file: %s\n",strerror(errno));
    exit(-4);
print_outfile_error:
    printf("Error opening output file: %s\n",strerror(errno));
    exit(-5);
}
