#include "adapcm.hpp"
#include "decode.hpp"
#include <cstdio>

using namespace ADAPCM;

uint16_t scale_used,coeff_used;
uint64_t coeff_histogram[16];

void decode_page_stereo(const byte *inbuffer, sample_t *outbuffer,uint samples_per_block) {
    constexpr bool stereo = true;

    const uint pagesamples = samples_per_page(stereo,samples_per_block);

    sample_t history2[2],history1[2];
    uint read_ptr = 0;
    uint block_cnt = 0;
    byte scale,coeff_set;
    bool joint_mode;

    for (uint si=0;si<pagesamples;si++) {
        sample_t decoded[2];
        if (si<2) { // first two samples are literal
            decoded[0] = inbuffer[read_ptr+0]+(inbuffer[read_ptr+1]<<8);
            decoded[1] = inbuffer[read_ptr+2]+(inbuffer[read_ptr+3]<<8);
            read_ptr += 4;
        } else {
            if (block_cnt == 0) { // new block
                byte blockhead = inbuffer[read_ptr++];
                byte sclnib = getnib0(blockhead);
                scale_used |= 1<<sclnib;
                if (sclnib <= 12) {
                    scale = sclnib;
                    joint_mode = false;
                } else {
                    scale = (sclnib-12)<<2;
                    joint_mode = true;
                }
                coeff_set = getnib1(blockhead);
                coeff_used |= 1<<coeff_set;
                coeff_histogram[coeff_set]++;
                block_cnt = samples_per_block;
            }
            block_cnt--;

            byte sample = inbuffer[read_ptr++];
            delta8_t delta[2];
            if (joint_mode) {
                delta[0] = delta[1] = delta8_t(sample);
            } else {
                delta[0] = delta8_t(getnib0(sample)<<4);
                delta[1] = delta8_t(getnib1(sample)<<4);
            }

            for (uint ch=0;ch<2;ch++) {
                int32_t prediction = predict_sample(coeff_set,history1[ch],history2[ch]);
                int32_t delta_scaled = (int32_t(delta[ch])<<scale)>>4;
                decoded[ch] = int32_to_sample_t(prediction+delta_scaled);
            }

        }
        for (uint ch=0;ch<2;ch++) {
            outbuffer[si*2+ch] = decoded[ch];
            history2[ch] = history1[ch];
            history1[ch] = decoded[ch];
        }

    }
    
}