#include "adapcm.hpp"
#include "encode.hpp"
#include <cstring>
#include <cstdio>

using namespace ADAPCM;

inline int32_t sar_round(int32_t v,uint shift) {
    //return (v+((1<<shift)>>1))/(1<<shift);
    return v>>shift; //Inexplicably less noise - might have to do with the -8..7 delta range
}

static void encode_block_stereo(const sample_t *inbuffer, byte *outbuffer,sample_t (&history1_outer)[2],sample_t (&history2_outer)[2],uint samples_per_block) {
    constexpr bool stereo = true;
    const uint blockbytes = bytes_per_block(stereo,samples_per_block);
    byte coeff_outdata[16][blockbytes];
    sample_t coeff_history[16][2][2];

    // Try every set of coefficients and pick the best one
    uint64_t min_error = UINT64_MAX;
    uint min_error_set = 0;
    for(uint coeff_set=0;coeff_set<16;coeff_set++) {
        sample_t (&history1)[2] = coeff_history[coeff_set][0];
        sample_t (&history2)[2] = coeff_history[coeff_set][1];
        byte (&outdata)[blockbytes] = coeff_outdata[coeff_set];
        uint64_t total_error;
        // Init history and guess scale
        /*
        uint32_t max_delta = 0;
        for (uint ch=0;ch<2;ch++) {
            uint32_t max_ch_delta = 0;
            history2[ch] = history2_outer[ch];
            history1[ch] = history1_outer[ch];
            for (uint si=0;si<samples_per_block;si++) {
                sample_t input = inbuffer[si*2+ch];
                int32_t prediction = predict_sample(coeff_set,inbuffer[(si-1)*2+ch],inbuffer[(si-2)*2+ch]);
                int32_t delta = input-prediction;
                max_ch_delta = std::max(max_ch_delta,(uint32_t)std::abs(delta));
                
            }
            max_delta = std::max(max_delta,max_ch_delta);
        }*/
        
        // Try encoding with increasing scale until it fits
        uint scale = 0;
        bool joint_mode = false;
        retry: {
            bool max_scale = scale == 12 || scale == 15;
            total_error = 0;

            bool overflow = false;
            bool try_joint_mode = !joint_mode;
            
            byte head = 0;
            setnib0(head,scale);
            setnib1(head,coeff_set);
            outdata[0] = head;
            // init history
            for (uint ch=0;ch<2;ch++) {
                history2[ch] = history2_outer[ch];
                history1[ch] = history1_outer[ch];
            }
            // Try encoding
            for (uint si=0;si<samples_per_block;si++) {
                int32_t mean_delta = 0;
                delta8_t delta8[2];
                int32_t prediction[2];
                for (uint ch=0;ch<2;ch++) {
                    int32_t target = int32_t(inbuffer[si*2+ch])-0x8000;
                    prediction[ch] = predict_sample(coeff_set,history1[ch],history2[ch]);
                    int32_t delta = target - prediction[ch];
                    if (joint_mode) {
                        mean_delta += delta;
                    } else {
                        auto scaled_delta = sar_round(delta,scale);
                        auto clamped_delta = std::clamp(scaled_delta,-8,7);
                        if (scaled_delta!=clamped_delta) overflow = true;
                        delta8[ch] = clamped_delta<<4;
                    }
                }
                mean_delta /= 2;
                if (joint_mode) {
                    auto scaled_delta = sar_round(mean_delta,((scale-13)<<2));
                    auto clamped_delta = std::clamp(scaled_delta,-128,127);
                    if (scaled_delta!=clamped_delta) overflow = true;
                    delta8[0] = delta8[1] = clamped_delta;
                    outdata[1+si] = byte(clamped_delta);
                } else {
                    if (delta8[0] != delta8[1]) try_joint_mode = false;
                    byte ib = 0;
                    setnib0(ib,delta8[0]>>4);
                    setnib1(ib,delta8[1]>>4);
                    outdata[1+si] = ib;
                }
                // decode compressed deltas again to produce accurate history data and detrmine error
                for (uint ch=0;ch<2;ch++) {
                    history2[ch] = history1[ch];
                    int32_t delta_scaled = (int32_t(delta8[ch])<<(joint_mode ? (scale-12)<<2 : scale))>>4;
                    history1[ch] = int32_to_sample_t(prediction[ch]+delta_scaled);
                    int64_t err = int32_t(history1[ch])-int32_t(inbuffer[si*2+ch]);
                    total_error += err*err;
                }
            }

            if (overflow && !max_scale) {
                scale++;
                goto retry;
            }
            if (try_joint_mode) {
                scale = joint_upgrade_tbl[scale];
                joint_mode = true;
                goto retry;
            }
        }
        if (total_error <= min_error) {
            min_error = total_error;
            min_error_set = coeff_set;
        }

    }

    //min_error_set = 5; //DELET THIS
    // Phew, done...
    memcpy(outbuffer,coeff_outdata[min_error_set],blockbytes);
    for (uint ch=0;ch<2;ch++) {
        history1_outer[ch] = coeff_history[min_error_set][0][ch];
        history2_outer[ch] = coeff_history[min_error_set][1][ch];
    }
}

void encode_page_stereo(const sample_t *inbuffer, byte *outbuffer,uint samples_per_block) {
    constexpr bool stereo = true;

    const uint pageblocks  = blocks_per_page(stereo,samples_per_block);

    sample_t history2[2],history1[2];
    uint read_ptr,write_ptr;

    // first two samples are literal
    history2[0] = inbuffer[0];
    history2[1] = inbuffer[1];
    history1[0] = inbuffer[2];
    history1[1] = inbuffer[3];

    outbuffer[0] = history2[0]&0xff;
    outbuffer[1] = history2[0]>>8;
    outbuffer[2] = history2[1]&0xff;
    outbuffer[3] = history2[1]>>8;
    outbuffer[4] = history1[0]&0xff;
    outbuffer[5] = history1[0]>>8;
    outbuffer[6] = history1[1]&0xff;
    outbuffer[7] = history1[1]>>8;

    write_ptr = 8;
    read_ptr = 4;

    // encode blocks
    for (uint block = 0;block<pageblocks;block++) {
        encode_block_stereo(inbuffer+read_ptr,outbuffer+write_ptr,history1,history2,samples_per_block);
        read_ptr += samples_per_block*2;
        write_ptr += bytes_per_block(stereo,samples_per_block);
    }

}